import asyncio
from typing import Any, Dict, Optional

from ..core import BaseClient

POST_PATH = "/users/{user_uuid}/creditreport"
GET_PATH = "/users/{user_uuid}/creditreport/{report_id}"


CreditReport = Dict[str, Any]


class BaseCreditReportResource:
    def __init__(self, client: BaseClient):
        self._client = client

    async def _create(self, user_uuid: str) -> str:
        async with self._client.session() as session:
            response = await session.post(POST_PATH.format(user_uuid=user_uuid))

        assert response.status_code == 201
        report_id = str(response.json()["credit_report_id"])
        return report_id

    async def _get_by_id(
        self, user_uuid: str, report_id: str
    ) -> Optional[CreditReport]:
        while True:
            async with self._client.session() as session:
                response = await session.get(
                    GET_PATH.format(user_uuid=user_uuid, report_id=report_id)
                )

            assert response.status_code == 200
            body: Dict[str, Any] = response.json()
            report_status = body["credit_report_status"]

            if report_status == "PENDING":
                await asyncio.sleep(5)
            elif report_status == "SUCCEEDED":
                report: CreditReport = body["credit_report"]
                return report
            else:
                return None

    async def _get(self, user_uuid: str) -> Optional[CreditReport]:
        report_id = await self._create(user_uuid)
        report = await self._get_by_id(user_uuid, report_id)
        return report


class AsyncCreditReportResource(BaseCreditReportResource):
    async def get(self, user_uuid: str) -> Optional[CreditReport]:
        return await self._get(user_uuid)


class SyncCreditReportResource(BaseCreditReportResource):
    def get(self, user_uuid: str) -> Optional[CreditReport]:
        return asyncio.run(self._get(user_uuid))



import asyncio
from abc import ABC
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from ..core import BaseClient, limit_concurrency
from ..encoders import encode_query_params

PATH = "/users"


class User(BaseModel):
    uuid: str
    first_name: str
    last_name: str
    email: str
    primary_phone_number: str
    primary_phone_imei: Optional[str]
    secondary_phone_number: Optional[str]
    secondary_phone_imei: Optional[str]
    is_kyc_verified: bool
    device_id: str
    external_id: Optional[str]
    created_at: datetime
    updated_at: datetime


class UsersMeta(BaseModel):
    client_uuid: str
    created_before: datetime
    created_after: datetime
    search: str
    total_users_count: int
    page: int
    max_pages: int


class UsersResponse(BaseModel):
    meta: UsersMeta
    users: List[User]

    class Config:
        fields = {"meta": "_meta"}


class BaseUsersResource(ABC):
    def __init__(self, client: BaseClient):
        self._client = client

    async def _get_page(self, search: Optional[str], page: int = 1) -> UsersResponse:
        async with self._client.session() as session:
            response = await session.get(
                PATH,
                params=encode_query_params(
                    search=search,
                    page=page,
                ),
            )

        assert response.status_code == 200, response.text
        return UsersResponse(**response.json())

    async def _get(self, search: Optional[str] = None) -> List[User]:
        response = await self._get_page(search)
        max_pages = response.meta.max_pages
        if max_pages > 1:
            coroutines = [
                self._get_page(search, page + 2) for page in range(max_pages - 1)
            ]
            response_pages = await asyncio.gather(
                *limit_concurrency(coroutines, self._client.concurrency_limit),
            )
            responses = (response, *response_pages)
        else:
            responses = (response,)

        return [user for response in responses for user in response.users]


class AsyncUsersResource(BaseUsersResource):
    async def get(self, search: Optional[str] = None) -> List[User]:
        return await self._get(search)


class SyncUsersResource(BaseUsersResource):
    def get(self, search: Optional[str] = None) -> List[User]:
        return asyncio.run(self._get(search))



import asyncio
from abc import ABC
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Coroutine, List, Sequence, TypeVar

import httpx

DEFAULT_BASE_URL = "https://api.pngme.com/beta"


T = TypeVar("T")


def limit_concurrency(
    coroutines: Sequence[Coroutine[None, None, T]], concurrency: int
) -> List[Coroutine[None, None, T]]:
    """Decorate coroutines to limit concurrency.

    Enforces a limit on the number of coroutines that can run concurrently in higher
    level asyncio-compatible concurrency managers like asyncio.gather(coroutines),
    asyncio.as_completed(coroutines), etc.
    """
    semaphore = asyncio.Semaphore(concurrency)

    async def with_concurrency_limit(coroutine: Coroutine[None, None, T]) -> T:
        async with semaphore:
            return await coroutine

    return [with_concurrency_limit(coroutine) for coroutine in coroutines]


class BaseClient(ABC):
    def __init__(
        self,
        access_token: str,
        concurrency_limit: int = 50,
        base_url: str = DEFAULT_BASE_URL,
    ):
        """Client SDK to interact with Pngme financial report resources.

        Args:
            access_token: API key from https://admin.pngme.com
            concurrency_limit: maximum concurrency for internal pagination handling
            base_url: root url for API resources
        """
        self.access_token = access_token
        self.concurrency_limit = concurrency_limit
        self.base_url = base_url

    @asynccontextmanager
    async def session(self) -> AsyncGenerator[httpx.AsyncClient, None]:
        """Configure a session with authorization and connection settings."""
        headers = {"Authorization": f"Bearer {self.access_token}"}
        transport = httpx.AsyncHTTPTransport(retries=10)
        async with httpx.AsyncClient(
            base_url=self.base_url,
            headers=headers,
            timeout=30,
            transport=transport,
        ) as session:
            yield session
